<?php
/**
 * BSS Commerce Co.
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the EULA
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at thisURL:
 * http://bsscommerce.com/Bss-Commerce-License.txt
 *
 * @category   BSS
 * @package    Bss_AutoInvoice
 * @author     Extension Team
 * @copyright  Copyright (c) 2017-2018 BSS Commerce Co. ( http://bsscommerce.com )
 * @license    http://bsscommerce.com/Bss-Commerce-License.txt
 */

namespace Bss\AutoInvoice\Helper;

use Magento\Framework\App\Helper\Context;
use Magento\Sales\Model\Order\Email\Sender\InvoiceSender;
use Magento\Sales\Model\Order\Email\Sender\ShipmentSender;
use Magento\Sales\Model\Service\InvoiceService;
use Magento\Sales\Model\Order\ShipmentDocumentFactory;
use Magento\Sales\Model\Order\ShipmentFactory;
use Magento\Sales\Model\Order;
use Magento\Framework\Exception\LocalizedException;
use Magento\Sales\Model\Order\Invoice;

/**
 * Class AutoInvoice
 *
 * @package Bss\AutoInvoice\Helper
 */
class AutoInvoice extends \Magento\Framework\App\Helper\AbstractHelper
{
    /**
     * @var Data
     */
    private $helper;

    /**
     * InvoiceSender
     *
     * @var InvoiceSender $invoiceSender
     */
    private $invoiceSender;

    /**
     * ShipmentSender
     *
     * @var ShipmentSender $shipmentSender
     */
    private $shipmentSender;

    /**
     * ShipmentFactory
     *
     * @var ShipmentFactory $shipmentFactory
     */
    private $shipmentFactory;

    /**
     * ShipmentDocumentFactory
     *
     * @var ShipmentDocumentFactory $shipmentDocumentFactory
     */
    private $shipmentDocumentFactory;

    /**
     * InvoiceService
     *
     * @var InvoiceService $invoiceService
     */
    private $invoiceService;

    /**
     * @var \Magento\Sales\Model\Convert\Order
     */
    private $orderConverter;

    /**
     * AutoInvoice constructor.
     * @param Context $context
     * @param InvoiceSender $invoiceSender
     * @param ShipmentSender $shipmentSender
     * @param ShipmentDocumentFactory $shipmentDocumentFactory
     * @param ShipmentFactory $shipmentFactory
     * @param InvoiceService $invoiceService
     * @param Data $helper
     * @param \Magento\Sales\Model\Convert\Order $orderConverter
     */
    public function __construct(
        Context $context,
        InvoiceSender $invoiceSender,
        ShipmentSender $shipmentSender,
        ShipmentDocumentFactory $shipmentDocumentFactory,
        ShipmentFactory $shipmentFactory,
        InvoiceService $invoiceService,
        Data $helper,
        \Magento\Sales\Model\Convert\Order $orderConverter
    ) {
        parent::__construct($context);
        $this->helper = $helper;
        $this->invoiceSender = $invoiceSender;
        $this->shipmentSender = $shipmentSender;
        $this->invoiceService = $invoiceService;
        $this->shipmentFactory = $shipmentFactory;
        $this->shipmentDocumentFactory = $shipmentDocumentFactory;
        $this->orderConverter = $orderConverter;
    }

    /**
     * @param Order $order
     * @return \Magento\Framework\DB\Transaction
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function createInvoice($order)
    {
        /** @var Invoice $invoice */
        $invoice = $this->invoiceService->prepareInvoice($order);
        $this->checkInvoice($invoice);
        if ($order->getPayment()->getMethodInstance()->isOffline()) {
            $invoice->setRequestedCaptureCase(Invoice::CAPTURE_OFFLINE);
        } else {
            $invoice->setRequestedCaptureCase($this->helper->getSelectCapture());
        }

        $transactionSave = false;

        try {
            $invoice->register();
            $invoice->getOrder()->setCustomerNoteNotify(false);
            $invoice->getOrder()->setIsInProcess(true);
            $transactionSave = $this->helper->returnTransaction()->create()->addObject(
                $invoice
            )->addObject(
                $invoice->getOrder()
            );
            $transactionSave->save();

            $order->addStatusHistoryComment(
                __('Automatically Invoiced by BSS AutoInvoice.'),
                false
            )->save();
            // send invoice email
            $this->invoiceSender->send($invoice);
        } catch (\Exception $e) {
            $this->helper->returnLogger()->error($e->getMessage());
        }
        return $transactionSave;
    }

    /**
     * @param Order $order
     */
    public function createShipment($order)
    {
        try {
            if (!$order->canShip()) {
                return;
            }

            $orderShipment = $this->orderConverter->toShipment($order);

            foreach ($order->getAllItems() as $orderItem) {
                // is virtual item and item qty valid
                if (!$orderItem->getQtyToShip() || $orderItem->getIsVirtual()) {
                    continue;
                }

                $qty = $orderItem->getQtyToShip();
                $shipmentItem = $this->orderConverter->itemToShipmentItem($orderItem)->setQty($qty);

                $orderShipment->addItem($shipmentItem);
            }

            $orderShipment->register();
            $orderShipment->getOrder()->setIsInProcess(true);

            // Save created Order Shipment
            $orderShipment->save();
            $orderShipment->getOrder()->save();

            // Send Shipment Email
            $this->sendShipment($orderShipment);
            if ($orderShipment->save()) {
                $order->addStatusHistoryComment(
                    __('Automatically Shipment by BSS AutoInvoice.'),
                    false
                )->save();
            }
        } catch (\Exception $e) {
            $this->helper->returnLogger()->error($e->getMessage());
        }
    }

    /**
     * Send Shipment
     *
     * @param mixed $shipment
     */
    protected function sendShipment($shipment)
    {
        if ($shipment) {
            try {
                $this->shipmentSender->send($shipment);
            } catch (\Exception $e) {
                $this->helper->returnLogger()->error($e->getMessage());
            }
        }
    }

    /**
     * @param Invoice $invoice
     * @throws LocalizedException
     */
    protected function checkInvoice($invoice)
    {
        if (!$invoice) {
            throw new LocalizedException(__('We can\'t save the invoice right now.'));
        }

        if (!$invoice->getTotalQty()) {
            throw new LocalizedException(
                __('You can\'t create an invoice without products.')
            );
        }
    }

    /**
     * Get Shipment By Version
     *
     * @return ShipmentDocumentFactory|ShipmentFactory
     */
    protected function getShipmentByVersion()
    {
        if ($this->helper->checkMagentoVersion()) {
            return $this->shipmentFactory;
        } else {
            return $this->shipmentDocumentFactory;
        }
    }
}
